# https://flask-sqlalchemy.palletsprojects.com/en/2.x/models/

# from flask import current_app
from store import db
from flask_login import UserMixin
from werkzeug.security import generate_password_hash, check_password_hash


# Helper tables for many to many relationships, see https://flask-sqlalchemy.palletsprojects.com/en/2.x/models/#many-to-many-relationships.
items = db.Table(
    "Items",
    db.Column("item_id", db.Integer, db.ForeignKey("item.id"), primary_key=True),
    db.Column(
        "repository_id", db.Integer, db.ForeignKey("repository.id"), primary_key=True
    ),
)

itemSets = db.Table(
    "ItemSets",
    db.Column("item_id", db.Integer, db.ForeignKey("item.id"), primary_key=True),
    db.Column(
        "item_set_id", db.Integer, db.ForeignKey("item_set.id"), primary_key=True
    ),
)


class Repository(db.Model):
    """The subset of items held by each different shop."""

    id = db.Column(db.Integer, primary_key=True)
    items = db.relationship(
        "Item",
        secondary=items,
        lazy="subquery",
        backref=db.backref("Repositories", lazy=True),
    )

    def __repr__(self):
        return f"id: {self.id}, items: {self.items}"


class Item(db.Model):
    """The master list of items for sale."""

    id = db.Column(db.Integer, primary_key=True)
    description = db.Column(db.String(256), nullable=False)
    price = db.Column(
        db.Integer, nullable=False
    )  # In pounds, as we're not dividing or multiplying this will not matter in calculations.
    date_sold = db.Column(db.Date, nullable=True)  # Null if unsold.

    def __repr__(self):
        return f"id: {self.id}, description: {self.description}, price: {self.price}, date_sold: {self.date_sold}"


class ItemSet(db.Model):
    """The subset of items in a collection."""

    id = db.Column(db.Integer, primary_key=True)
    description = db.Column(db.String(256), nullable=False)
    items = db.relationship(
        "Item",
        secondary=itemSets,
        lazy="subquery",
        backref=db.backref("ItemSets", lazy=True),
    )

    @property
    def price(self):
        return sum(int(item.price) for item in self.items)

    @property
    def quantity(self):
        return len(self.items)

    def __repr__(self):
        return f"id: {self.id}, description: {self.description}, items: {self.items}, quantity: {self.price}, price: { self.price() }"  # TODO: Rename price to quantity


class User(db.Model, UserMixin):
    """database for user register it will be added more cullomns in the future
    that is a basic implementation to check if it works
    Freddy implemented that
    UserMixin is to inherit flask-login implementetion"""

    user_id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    username = db.Column(db.String(256), nullable=False, unique=True)
    email = db.Column(db.String(256), nullable=False)
    phone_number = db.Column(db.String(20), nullable=False)
    # The hashing algorithm determines how much storage is taken in the databasse,
    # as the password itself can be of a theoretically infinite length but hashed
    # into a sepcific size.
    password_hash = db.Column(db.String(256), nullable=False)
    userType = db.Column(db.String(20), default="standard")

    @classmethod
    def create_user(
        cls,
        username: str,
        password: str,
        email: str,
        phone_number: str,
    ):
        password_hash = generate_password_hash(password)
        user = cls(
            username=username,
            password_hash=password_hash,
            email=email,
            phone_number=phone_number,
        )
        db.session.add(user)
        db.session.commit()
        return user

    @classmethod
    def update_passwords(cls, user_id: int, password: str):
        user = cls.query.get(user_id)
        password_hash = generate_password_hash(password)
        user.password_hash = password_hash
        db.session.commit()

    @classmethod
    def update_username(cls, user_id: int, username: str):
        user = cls.query.get(user_id)
        user.username = username
        db.session.commit()

    @classmethod
    def update_email(cls, user_id: int, email: str):
        user = cls.query.get(user_id)
        user.email = email
        db.session.commit()

    @classmethod
    def update_Number(cls, user_id: int, phone_number: str):
        user = cls.query.get(user_id)
        user.email = phone_number
        db.session.commit()

    def is_active(self) -> bool:
        return True

    def get_userDetails(self):
        lis = [self.username, self.email, self.phone_number]
        return lis

    def get_id(self):
        return str(self.user_id)

    def check_password(self, password: str) -> bool:
        return check_password_hash(self.password_hash, password)

    def checkIfUserExist(self, username: str) -> bool:
        if self.username == username:
            return True
        else:
            return False

    def __repr__(self) -> str:
        return f"User(user_id={self.user_id}, username='{self.username}')"


class Adresses(db.Model):
    order_id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(db.Integer)
    country = db.Column(db.String(256), nullable=False)
    homeNumber = db.Column(db.String(256), nullable=False, unique=True)
    streetName = db.Column(db.String(256), nullable=False)
    city = db.Column(db.String(30), nullable=False)
    phoneNumber = db.Column(db.String(30), nullable=False)
    contactEmail = db.Column(db.String(50), nullable=False)

    @classmethod
    def create_Adresses(
        cls,
        user_id: int,
        country: str,
        homeNumber: str,
        streetName: str,
        city: str,
        phoneNumber: str,
        contactEmail: str,
    ):
        adresses = cls(
            user_id=user_id,
            country=country,
            homeNumber=homeNumber,
            streetName=streetName,
            city=city,
            phoneNumber=phoneNumber,
            contactEmail=contactEmail,
        )
        db.session.add(adresses)
        db.session.commit()
        return adresses
