# TODO: Search for similiarly named item to replace removed item from itemset.

from difflib import SequenceMatcher
from wtforms import ValidationError
from store.models import *
from sqlalchemy import select, func
import datetime
from store import *
import re
import requests


# Difference between filter and filter_by: https://stackoverflow.com/a/31560897
def get_unsold_items():
    return Item.query.filter_by(date_sold=None).all()


def get_sold_items():
    return Item.query.filter(Item.date_sold.isnot(None)).all()


def get_items():
    return Item.query.all()


def get_item_by_id(item_id):
    item = Item.query.get(item_id)
    if item is not None:
        return item
    else:
        # TODO error handling
        # TODO error handling
        pass


def get_item_sets():
    return ItemSet.query.all()


def get_item_set_by_id(itemset_id):
    itemset = ItemSet.query.get(itemset_id)
    if itemset != None:
        return itemset
    else:
        # TODO error handling
        pass


def get_similar_items(description):
    """Search the description of all items to find similar.
    This is intended for use in replacing items in an items,
    and works best when description is close to the target.
    As a result it does not return sold items.
    TODO: Do the regex using SQL alchemy so that as much as
    possible is handled by the database rather than in memory
    for maximum performance.
    """
    to_search = get_unsold_items()
    found = []
    for descriptable in to_search:
        # Remove date from description to make comparisons more accurate. Down the line might be worth creating a list of phrases to exclude.
        description_to_compare = descriptable.description.lower()
        description_to_compare = re.sub("[0-9]*th century", "", description_to_compare)
        description_to_compare = re.sub(
            "[0-9][0-9][0-9][0-9]", "", description_to_compare
        )

        # This relies on the fact both Item and Items sets have a description column.
        if (
            SequenceMatcher(
                None, description.lower(), descriptable.description.lower()
            ).ratio()
            > 0.5  # This ratio comparison needs to be fine tuned.
            and descriptable.date_sold == None
        ):
            found.append(descriptable)

    return found


def get_by_keyword(substring):
    """Search desciption for matching substring. Case insensitive.
    This is most useful for a user keyword search.
    """
    substring = substring.lower()
    items = Item.query.filter(func.lower(Item.description.contains(substring))).all()
    items = [item for item in items if item.date_sold == None]
    item_sets = ItemSet.query.filter(
        func.lower(ItemSet.description.contains(substring))
    ).all()

    return item_sets + items


def get_itemsets_by_item_id(item_id):
    item_sets = get_item_sets()
    item_set_contained_id = []
    for item_set in item_sets:
        for item in item_set.items:
            if item.id == item_id:
                item_set_contained_id.append(item_set)
    return item_set_contained_id


def get_item_by_id(id):
    return Item.query.get(id)


def get_items_by_itemset_id(item_set_id):
    item_set = get_item_set_by_id(item_set_id)
    items = item_set.items
    return items


def add_item(description, price, date_sold=None):
    item = Item(description=description, price=price, date_sold=date_sold)
    db.session.add(item)
    db.session.commit()


def remove_item(item):
    update_set(item)
    db.session.delete(item)
    db.session.commit()


def remove_item_set(item):
    db.session.delete(item)
    db.session.commit()


def sell_item_set(item_set, date=datetime.datetime.today()):
    for item in item_set.items:
        item.date_sold = date

    db.session.commit()
    remove_item_set(item_set)
    db.session.commit()


def sell_item(item, date=datetime.datetime.today()):
    update_set(item)
    item.date_sold = date
    db.session.commit()


def update_set(reference_item):
    """Update the set that the passed item belongs to by removing it from the set
    and adding a similar item."""

    # We need the owning item set to avoid adding duplicates.
    owning_itemsets = get_itemsets_by_item_id(reference_item.id)
    # Sort by closest match. Could also do it based of the description of the set.
    similar_items = get_similar_items(reference_item.description)
    similar_items.sort(
        key=lambda x: SequenceMatcher(
            None, x.description, reference_item.description
        ).ratio()
    )
    similar_items.reverse()

    # Remove items already in set from sorted items
    for item_set in owning_itemsets:
        for item in item_set.items:
            if item in similar_items:
                similar_items.remove(item)

    # Remove items already in the itemset, and add new one.
    # Asking the tutors they say item sets can share items,
    # so may as well use the first element in sorted_items.
    for itemset in owning_itemsets:
        itemset.items.remove(reference_item)
        itemset.items.append(similar_items[0])
    db.session.commit()


def get_item_by_id(id):
    return Item.query.get(id)
